/**
 * \file
 *         Implmentation of Compact-LWE puiblic key encryption scheme developed in CSIRO
 *         09-11-2017
 *         Dongxi Liu <dongxi.liu@csiro.au>
 */

#ifndef clwe_crypt_h
#define clwe_crypt_h

#include "mini-gmp.h"
#include "api.h"
#define M 128u  
#define N 9u  
#define T_BYTES 4u 
#define Q_BYTES 8u
#define W_POS 224u
#define W_NEG 32u   
#define pSIZE "16777216"
#define eMax  "3200"
#define F 2u
#define B "68719476736" //value of b' in specification
#define L 5 //3 is hard coded; L=5 means 8 in specification
//parameter b hard coded; parameter eMin is derived in implementaiton 
//#define debug


struct Cipher{
        mpz_t head[N-1];        
        mpz_t tail[F]; 
	unsigned char cval[T_BYTES]; 
};


struct SK{
	unsigned char n;
	unsigned char m;
	unsigned char q_bytes;
	unsigned char t_bytes;
	mpz_t wPos;
	mpz_t wNeg;
	mpz_t p;
        mpz_t q;
	mpz_t r;
	mpz_t t;
	mpz_t ck[F];
	mpz_t ckInv;
	mpz_t s[F][N];
        mpz_t k[F];
        mpz_t kInv[F];
	mpz_t bv; //corresponding to b' specification
};

struct PK{
        mpz_t element[F][M];   
	mpz_t q;		
	mpz_t t;
        mpz_t wPos;	
	mpz_t wNeg;
	mpz_t bv;
	unsigned char m;
	unsigned char n;
	unsigned char q_bytes;
	unsigned char t_bytes;
};

//I is an array of 256 random bytes; no specific requirements
static const unsigned char I[] = {
 0x83, 0x17, 0xe3, 0x6e, 0x5c, 0xc5, 0x8a, 0xda, 0xcb, 0x8c, 0x98, 0x25, 0xdf, 0x77, 0x8d, 0xda,
 0x2a, 0x91, 0x99, 0x92, 0x77, 0x1b, 0x5a, 0x73, 0xf7, 0xcf, 0x2a, 0xa9, 0xe3, 0xd7, 0x2d, 0xc7, 
 0x7f, 0x5b, 0x3e, 0x25, 0x42, 0x3a, 0xfe, 0x5a, 0x2b, 0x74, 0xdc, 0x65, 0x3c, 0xa3, 0x85, 0xc1,
 0x49, 0xab, 0xb7, 0x80, 0xf4, 0x8a, 0x2d, 0x89, 0xdb, 0xed, 0x39, 0xcc, 0x9c, 0x31, 0x16, 0xd4,
 0xd4, 0xa5, 0x6b, 0xf9, 0x29, 0xa4, 0x45, 0x2c, 0x68, 0xd1, 0x3a, 0x38, 0x5a, 0x7f, 0x35, 0x42, 
 0x96, 0x15, 0xbf, 0x92, 0xc4, 0x4d, 0xc1, 0xcf, 0x33, 0x7c, 0x24, 0xbd, 0x98, 0x7d, 0x2e, 0x8b,
 0xb1, 0x0b, 0xf6, 0x75, 0x76, 0xcc, 0x47, 0x21, 0x79, 0x5c, 0x7f, 0x5f, 0x36, 0x73, 0x1a, 0x7a,
 0xeb, 0x55, 0x3d, 0x69, 0x5f, 0xfa, 0x4e, 0x21, 0x87, 0xc4, 0xbc, 0xab, 0xfb, 0x32, 0x08, 0x39,
 0xdb, 0x69, 0x43, 0x64, 0xfc, 0x46, 0x2a, 0xdd, 0x1f, 0x21, 0x41, 0xdd, 0x39, 0x4f, 0x67, 0x28,
 0x7b, 0x7c, 0x5b, 0x87, 0x1e, 0x31, 0x11, 0x58, 0x5a, 0x19, 0x77, 0x47, 0xee, 0x13, 0xb7, 0x55,
 0x7a, 0x7e, 0x1c, 0x26, 0xa3, 0xb7, 0x9b, 0xc1, 0x9e, 0x17, 0xc4, 0x03, 0xc1, 0xeb, 0xa6, 0x2d,
 0x5f, 0x49, 0x73, 0x01, 0xae, 0x61, 0x39, 0x9a, 0xfb, 0xa1, 0x8d, 0x53, 0x92, 0xa2, 0x71, 0x49,
 0x77, 0x8f, 0xd5, 0xe4, 0x0e, 0x07, 0xaf, 0x97, 0xe8, 0xdf, 0x0a, 0x2d, 0xb5, 0x29, 0x36, 0x8a,
 0x72, 0x67, 0xcf, 0xbe, 0x02, 0x66, 0x52, 0x23, 0x05, 0xd3, 0xe7, 0x83, 0x02, 0x2b, 0x9e, 0x58,
 0x51, 0x78, 0x05, 0x81, 0xce, 0x0d, 0x45, 0xb8, 0x93, 0xbf, 0xd7, 0xca, 0x1f, 0x1b, 0x24, 0x5f,
 0xe5, 0x14, 0x12, 0xde, 0x4b, 0x6b, 0x9d, 0x11, 0x69, 0x67, 0x63, 0x4e, 0x53, 0x67, 0x09, 0xdc,
};



int sk_gen(struct SK* sk);
int pk_gen(struct SK* sk, unsigned char *pk_str);
int enc(struct PK* pk, const unsigned char* v,struct Cipher*); 
int cipher_out_str(struct Cipher* D, unsigned char* buf, int buf_len);
void cipher_in_str(const unsigned char* buf, int* last, struct Cipher* D);
int pk_out_str(struct PK* pk, unsigned char* pk_str, int buf_len);
struct PK* pk_in_str(const unsigned char* pk_str, int len);
int sk_out_str(struct SK* sk, unsigned char* buf, int buf_len);
struct SK* sk_in_str(const unsigned char* buf, int buf_len);
void clearCipher(struct Cipher* D);
int dec(struct SK* sk, struct Cipher*, unsigned char* v);
void randomMPZ(int len, mpz_t a);
int encode(const unsigned char* buf, unsigned long long buf_len, unsigned char* res, unsigned long long res_len);
int decode(unsigned char* buf, unsigned long long* buf_len, const unsigned char* res, unsigned long long res_len);
void clearPK(struct PK* pk);
#endif
